import datetime
from pqsdk.interface import AbstractOrder
from pqsdk.enums.orderStatus import OrderStatus


class Order(AbstractOrder):

    def __init__(self,
                 order_id: int,
                 stock_code: str,
                 direction: str,
                 order_volume: int,
                 order_price: float,  # 委托价格
                 avg_cost: float,  # 平均成本
                 context):
        self._order_id = order_id
        self.stock_code = stock_code
        self.direction = direction
        self.order_volume = order_volume
        self.order_price = order_price  # 委托价格
        # 平均成本：卖出时表示下卖单前的此股票的持仓成本, 用来计算此次卖出的收益. 买入时表示此次买入的均价(等同于price).
        self._avg_cost = avg_cost
        self.context = context

    @property
    def status(self) -> OrderStatus:
        return OrderStatus.held

    @property
    def add_time(self) -> datetime.datetime:
        return self.context.current_dt

    @property
    def is_buy(self) -> bool:
        return True if self.direction == 'buy' else False

    @property
    def volume(self):
        """
        下单数量, 不管是买还是卖, 都是正数
        :return:
        """
        return self.order_volume

    @property
    def filled(self):
        """
        已经成交的股票数量, 正数
        :return:
        """
        return self.order_volume

    @property
    def security(self):
        return self.stock_code

    @property
    def order_id(self):
        return self._order_id

    @property
    def price(self):
        """
        平均成交价格, 已经成交的股票的平均成交价格(一个订单可能分多次成交)
        :return:
        """
        return self.order_price

    @property
    def avg_cost(self):
        """
        卖出时表示下卖单前的此股票的持仓成本, 用来计算此次卖出的收益. 买入时表示此次买入的均价(等同于price).
        :return:
        """
        return self._avg_cost

    @classmethod
    def trade_fee_calculation(cls, exchange, price, volume, commission_rate, trade_type):
        """
        计算股票交易手续费
        :param exchange: 交易所，SH 代表上海A股；SZ 代表深圳A股
        :param price: 股票成交价格
        :param volume: 股票成交量
        :param commission_rate: 佣金比率
        :param trade_type: 交易类型，B代表买入，S代表卖出
        :return: total_fee 费用总计
        """
        stamp_duty, transfer_fee, exchange_fee, manage_fee, commission = 0, 0, 0, 0, 0
        if trade_type == 'S':
            # 印花税
            stamp_duty = price * volume * 0.001
        if exchange == 'SH':
            # 过户费 最低1元
            transfer_fee = max(price * volume * 0.00001, 1)
        # 经手费
        exchange_fee = price * volume * 0.0000487
        # 证管费
        manage_fee = price * volume * 0.00002
        # 佣金 最低5元
        commission = max(price * volume * commission_rate, 5)
        total_fee = stamp_duty + transfer_fee + exchange_fee + manage_fee + commission
        return total_fee

    @property
    def commission(self):
        """
        交易费用（佣金、税费等）
        :return:
        """
        ex = self.security[-2:]
        return self.trade_fee_calculation(exchange=ex,
                                          price=self.price,
                                          volume=self.filled,
                                          commission_rate=self.context.commission,
                                          trade_type='B' if self.is_buy else 'S')



from collections.abc import MutableMapping
from collections import OrderedDict
from pqsdk.interface import AbstractRunInfo
from .instrument import Instrument
from pqsdk.api import get_index_members_lst


class CurrentDataCacheDict(MutableMapping):
    """
    实现字典里的数据按需获取的功能，即当读取字典里的数据的时候，再去数据库中读取数据,
    并且考虑添加缓存机制来提升性能，考虑限制缓存的大小、使用LRU算法等方式进行优化.
    """

    def __init__(self, context, run_info: AbstractRunInfo, max_size=1000):
        self.context = context
        self.run_info = run_info
        self.max_size = max_size
        self._cache = OrderedDict()

    def __getitem__(self, key):
        """
        根据股票代码的key获取股票详细信息
        :param key:
        :return:
        """
        if key in self._cache:
            # 如果key在缓存中，则将其移到最前面
            value = self._cache.pop(key)
            self._cache[key] = value
        else:
            # 如果key不在缓存中，则从数据库中读取值，并将其添加到缓存中
            value = Instrument(stock_code=key, run_info=self.run_info, context=self.context)

            if value.name is None:
                # 如果查询结果为空，则抛出KeyError异常
                raise KeyError(key)
            else:
                # 将查询结果添加到缓存中
                self._cache[key] = value
                self._enforce_max_size()

        return value

    def __setitem__(self, key, value):
        """
        将键值对添加到缓存
        :param key:
        :param value:
        :return:
        """
        if key in self._cache:
            # 如果key在缓存中，则将其移到最前面
            self._cache.pop(key)
        self._cache[key] = value

        self._enforce_max_size()

    def __delitem__(self, key):
        """
        从缓存中删除键值对
        :param key:
        :return:
        """
        del self._cache[key]

    def __iter__(self):
        """
        遍历所有键
        :return:
        """
        # 从回测的股票池的成分股获取股票列表
        stock_lst = get_index_members_lst(self.context.stock_pool,
                                          trade_date=self.context.current_dt.strftime('%Y-%m-%d'))

        for stock_code in stock_lst:
            yield Instrument(stock_code=stock_code, run_info=self.run_info, context=self.context)

    def __len__(self):
        """
        获取键的总数
        :return:
        """
        # 从回测的股票池的成分股获取股票列表
        stock_lst = get_index_members_lst(self.context.stock_pool,
                                          trade_date=self.context.current_dt.strftime('%Y-%m-%d'))

        return len(stock_lst)

    def _enforce_max_size(self):
        """
        如果缓存大小超过了max_size，则删除最旧的元素
        :return:
        """
        while len(self._cache) > self.max_size:
            self._cache.popitem(last=False)



from six import with_metaclass
import abc
from typing import Callable, Union, List
from .abstractPosition import AbstractPosition
from .abstractInstrument import AbstractInstrument
import datetime
import pandas as pd


class AbstractStrategyContext(with_metaclass(abc.ABCMeta)):
    """
    策略上下文的抽象接口类。
    """

    @abc.abstractmethod
    def plot_line(self, sec_code: str, name: str, value: float, desc=None):
        """
        画曲线
        :param sec_code:
        :param name:
        :param value:
        :param desc:
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def plot_bar(self, sec_code: str, name: str, value: float, desc=None):
        """
        画柱状图
        :param sec_code:
        :param name:
        :param value:
        :param desc:
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def run_monthly(self, func: Callable, monthday: int, time_str: str) -> None:
        """
        按月运行

        :param func: 一个自定义的函数, 此函数必须接受context参数;例如自定义函数名market_open(context)
        :param monthday: 每月的第几个交易日, day (1-31)
        :param time_str: 具体执行时间,一个字符串格式的时间, 24小时内的任意时间，例如"10:00", "01:00"；
        """
        raise NotImplementedError

    @abc.abstractmethod
    def run_weekly(self, func: Callable, weekday: int, time_str: str) -> None:
        """
        按周运行

        :param func: 一个自定义的函数, 此函数必须接受context参数;例如自定义函数名market_open(context)
        :param weekday: 每周的第几个交易日, 1 = monday, ... 7 = sunday
        :param time_str: 具体执行时间,一个字符串格式的时间, 24小时内的任意时间，例如"10:00", "01:00"；
        """
        raise NotImplementedError

    @abc.abstractmethod
    def run_periodically(self,
                         func: Callable,
                         days: int,
                         time_str: str,
                         start_date: str = None,
                         end_date: str = None) -> None:
        """
        按调仓周期运行，如果遇到调仓日为非交易日，则顺延到下一个交易日。

        :param func: 一个自定义的函数, 此函数必须接受context参数;例如自定义函数名market_open(context)
        :param days : 间隔天数执行，start_date为第一次执行日期
        :param time_str: 具体执行时间,一个字符串格式的时间, 24小时内的任意时间，例如"10:00", "01:00"；在实盘中才生效
        :param start_date: 开始日期，默认为回测开始日期
        :param end_date: 结束日期, 默认为回测结束日期
        """
        raise NotImplementedError

    @abc.abstractmethod
    def run_daily(self, func: Callable, time_str: str) -> None:
        """
        每天内何时运行

        :param func: 一个自定义的函数, 此函数必须接受context参数;例如自定义函数名market_open(context)
        :param time_str: 具体执行时间,一个字符串格式的时间, 24小时内的任意时间，例如"10:00", "01:00"；
        """
        raise NotImplementedError

    @abc.abstractmethod
    def run_minutely(self, func: Callable, time_str: str, minutes: int = 1) -> None:
        """
        每天内何时开始，间隔n分钟执行

        :param func: 一个自定义的函数, 此函数必须接受context参数;例如自定义函数名market_open(context)
        :param time_str: 具体执行时间,一个字符串格式的时间, 24小时内的任意时间，例如"10:00", "01:00"；
        :param minutes: 间隔执行的分钟数
        """
        raise NotImplementedError

    @abc.abstractmethod
    def run_secondly(self, func: Callable, time_str: str, seconds: int = 1) -> None:
        """
        每天内何时开始，间隔n秒执行

        :param func: 一个自定义的函数, 此函数必须接受context参数;例如自定义函数名market_open(context)
        :param time_str: 具体执行时间,一个字符串格式的时间, 24小时内的任意时间，例如"10:00", "01:00"；
        :param seconds: 间隔执行的秒数
        """
        raise NotImplementedError

    @abc.abstractmethod
    def cancel(self, order, force: bool = False):
        """
        取消订单委托
        :param force: 强制交易
        :param order: 回测中为order对象，实盘中为order_id
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def close(self, stock_code: str, force: bool = False):
        """
        以最新价平仓股票代码, 默认平仓数量为可用数量
        :param force: 强制交易
        :param stock_code: 需要平仓的股票代码
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_current_data(self):
        """
        获取当前单位时间（当天/当前分钟）的涨跌停价, 是否停牌，当天的开盘价等

        :return: 一个dict, 其中 key 是股票代码, value 是拥有如下属性的对象, 返回的结果只在当天有效:
                last_price : 最新价
                high_limit: 涨停价
                low_limit: 跌停价
                paused: 是否停止或者暂停了交易, 当停牌、未上市或者退市后返回 True
                is_st: 是否是 ST(包括ST, *ST)，是则返回 True，否则返回 False
                day_open: 当天开盘价
                name: 股票现在的名称, 可以用这个来判断股票当天是否是 ST, *ST, 是否快要退市
                industry_code: 股票现在所属行业代码
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_position(self, stock_code) -> AbstractPosition:
        """
        获取已经登录账号的股票持仓，返回个股持仓字典

        :param stock_code:
        :return: {'volume': 72000, 'can_use_volume': 0, 'open_price': 10.8, 'market_value': 1020240.0}
        """
        raise NotImplementedError

    @abc.abstractmethod
    def order_target_value(self,
                           stock_code: str,
                           target: float = 0.0,
                           price: float = None,
                           force: bool = False):
        """
        Place an order to rebalance a position to have final value of
        ``target``

        The current ``value`` is taken into account as the start point to
        achieve ``target``

          - If no ``target`` then close postion on data
          - If ``target`` > ``value`` then buy on data
          - If ``target`` < ``value`` then sell on data

        It returns either:

          - The generated order

          or

          - ``None`` if no order has been issued

        :param force: 强制交易
        :param stock_code: 目标持仓股票
        :param target: 目标持仓市值
        :param price:
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def order_target_percent(self,
                             stock_code: str,
                             target: float = 0.0,
                             force: bool = False):
        """
        基于order_target_value（），按照总资产的百分比委托下单

        :param force: 强制交易
        :param stock_code:目标持仓股票
        :param target: 目标持仓比例
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def order_target_volume(self,
                            stock_code: str,
                            target: float = 0.0,
                            force: bool = False):
        """
        基于order_target_value()，按照个股的数量目标数量下单。注意，可能存在账户资金不足的情况。

        :param force: 强制交易
        :param stock_code: 目标持仓股票
        :param target: 目标持仓数量
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def constant(self):
        """
        上下文的系统常量
        :return:
        """
        raise NotImplementedError

    # @property
    # @abc.abstractmethod
    # def planned_trade_dates(self) -> list:
    #     """
    #     获取调仓交易日列表。
    #     :return:
    #     """
    #     raise NotImplementedError

    @property
    @abc.abstractmethod
    def previous_trade_date(self) -> datetime.date:
        """
        当前bar对应的前一个交易日
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def parameters(self) -> dict:
        """
        执行策略的参数字典
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def stock_pool(self):
        """
        股票池
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def commission(self):
        """
        佣金费率
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def strategy_name(self):
        """
        策略名称
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def unit(self):
        """
        unit: 单位时间长度，支持1d、1m，默认为1d
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def dividend_type(self):
        """
        dividend_type: 复权选项(对股票/基金的价格字段、成交量字段及factor字段生效)
                'front'
                : 前复权, 默认是前复权
                none
                : 不复权, 返回实际价格
                'back'
                : 后复权
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def current_dt(self) -> datetime.datetime:
        """
        当前bar对应的日期/时间, format: %Y-%m-%d %H:%M:%S
        :return:
        """
        raise NotImplementedError

    # @property
    # @abc.abstractmethod
    # def next_trade_datetime(self):
    #     """
    #     当前bar对应的下一个日期/时间, format: %Y-%m-%d %H:%M:%S
    #     :return:
    #     """
    #     raise NotImplementedError

    @abc.abstractmethod
    def buy(self,
            stock_code: str,
            volume: int,
            price_type: int,
            price: float,
            force: bool = False):
        """
        证券买入委托
        Returns:
          - 委托的order_id
        """
        raise NotImplementedError

    @abc.abstractmethod
    def sell(self,
             stock_code: str,
             volume: int,
             price_type: int,
             price: float,
             force: bool = False):
        """
        证券卖出委托
        Returns:
          - 委托的order_id
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def portfolio(self):
        """
        策略投资组合，可通过该对象获取当前策略账户、持仓等信息
        """
        raise NotImplementedError

    @abc.abstractmethod
    def inout_cash(self, cash: float):
        """
        投资组合转入或转出资金，当日的出入金从当日开始记入成本，用于计算收益，即当日结束计算收益时的本金是包含当日出入金金额的
        :param cash: 可正可负，正为入金，负为出金。
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def run_info(self):
        """
        策略运行信息，包括回测过程中的所有参数。
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_orders(self, trade_date: str = None, is_buy: bool = None, include_canceled: bool = True):
        """
        获取交易日的委托单，返回DataFrame

        :param include_canceled: 是否包括已撤的委托
        :param trade_date:
        :param is_buy:
        :return: DataFrame,
            字段：['order_code', 委托编号
                'sec_code',  股票代码
                'order_time', 委托时间
                'is_buy', 委托买卖方向：买单=1， 卖单=0
                'order_volume',  委托数量
                'order_price']  委托价格
        """
        raise NotImplementedError



from six import with_metaclass
import abc


class AbstractPortfolio(with_metaclass(abc.ABCMeta)):
    """
    策略账户持仓信息
    """

    # @property
    # @abc.abstractmethod
    # def asset(self) -> dict:
    #     """
    #     账号资金
    #     {'total_asset': 10204239.17, 'cash': 9136261.17, 'market_value': 1067978.0, 'frozen_cash': 0.0}
    #     :return:
    #     """
    #     raise NotImplementedError

    @property
    @abc.abstractmethod
    def positions(self) -> dict:
        """
        [dict] 持仓字典
        {
        "510300.SH": {
          "volume": 2600,
          "can_use_volume": 1400,
          "open_price": 20.206,
          "market_value": 10701.599999999999
        }}
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def total_value(self):
        """
        [float]总的权益, 包括现金, 仓位的总价值, 可用来计算收益
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def positions_value(self):
        """
        [float] 持仓价值
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def available_cash(self):
        """
        [float] 可用资金
        """
        raise NotImplementedError

    @abc.abstractmethod
    def add_cash(self, cash: float):
        """
        添加/减少可用金额
        :param cash:
        :return:
        """
        raise NotImplementedError

    def get_frozen_cash(self, sec_code: str = None):
        """
        获取冻结现金
        冻结现金 = sum（（买入委托数量 - 买入成交数量） * 委托价格）

        在回测过程中，买入委托即刻成交，所以冻结现金默认为0

        :param sec_code: 获取指定股票的冻结现金, 如果为None，则获取投资组合的冻结现金
        :return:
        """
        raise NotImplementedError



# coding=utf-8
import datetime
import time
import pqsdk.api as api
from pqsdk.utils.dynamic_import import check_module, import_module_from_spec, import_module_from_code
from pqsdk.utils.import_global_modules import import_modules
from pqsdk import log
from .context import StrategyContext
from pqsdk.utils.timer_factory import TimerFactory
import pandas as pd
from .analyzer import TimeReturn
from typing import Callable
import pqsdk.utils.file_util as fu
from tqdm import tqdm


# 初始化策略前，导入sdk中的所有对象、函数、属性到全局变量中，以支持在策略中调用
import_modules(api)


class BacktestExecutor:
    def __init__(self, kwargs: dict, task=None):
        """

        :param kwargs:
        :param task: celery task
        """

        self.task = task

        # 策略参数, 从list转换为dict， 根据数据推断，把value转换为正确的数据类型
        # self.strategy_params = {param['key']: convert_to_type(param['value']) for param in kwargs.get('parameters')}
        self.strategy_params = kwargs.get('parameters')

        # # 检查必须的输入参数
        must_have_params = {'benchmark': "行情基准",
                            'unit': "行情周期",
                            "cash": "回测初始资金",
                            'start_date': "回测开始日期",
                            'end_date': "回测结束日期",
                            }
        for param, desc in must_have_params.items():
            if param not in self.strategy_params:
                content = f"输入参数中缺少必须的参数：{param}: {desc}"
                raise Exception(content)

        # 默认回测参数
        self.params = dict(
            # 指数 [000300.SH,000905.SH,000852.SH]
            index=self.strategy_params.get('index', []),
            # 股票池
            stock_pool=self.strategy_params.get('stock_pool', []),
            # 行情基准-运行周期，支持1d，1m，5m，即根据行情基准-证券代码的k线图，按照1d,1m执行handle_bar()函数
            unit=self.strategy_params.get('unit', '1d'),
            # 除权方式，, 支持none：不复权，front：前复权，back：后复权
            dividend_type=self.strategy_params.get('dividend_type', 'front'),  # 默认前复权
            strategy_file=kwargs.get('strategy_file', None),  # 策略代码文件
            strategy_script=kwargs.get('script', None),  # 策略代码
            parameters=self.strategy_params,  # dict类型，策略中可以访问到到自定义参数列表
            adjust_period=self.strategy_params.get('adjust_period', 5),  # 调仓周期，结合start_date和end_date计算调仓日列表
            hold_maxsize=self.strategy_params.get('hold_maxsize', 10),  # 最大持仓股票数量
            start_date=self.strategy_params.get('start_date'),  # 回测开始日期
            end_date=self.strategy_params.get('end_date'),  # 回测结束日期
            excluded_dates=None,  # 排除不交易的日期
            benchmark=self.strategy_params.get('benchmark', '000300.SH'),  # 回测基准
            init_investment=self.strategy_params.get('cash', 1000000),  # 初始回测资金
            commission=0.0,  # 交易佣金费率
            slip_type="perc",
            slip_perc=0.0,
            slip_fixed=0.0,
            print_dev=True,  # 是否打印开发日志
            save_result=True,  # 是否保存回测结果
            save_path="storage",
            save_tearsheet=True,  # 保存Tear sheet
            save_db=False,
        )

        # 初始化策略程序
        if self.params['strategy_file']:
            module_spec = check_module(self.params['strategy_file'])
            if module_spec:
                self.strategy = import_module_from_spec(module_spec)
                log.debug(f"从文件获取策略对象：strat_path = {self.params['strategy_file']}")
        elif self.params['strategy_script']:
            self.strategy = import_module_from_code(code=self.params['strategy_script'])
            log.info(f"从代码获取策略程序")
        else:
            raise Exception("未找到策略代码，中止回测程序")

        # 自定义定时器工厂
        self.timer_factory = TimerFactory(unit=self.params['unit'])

        # 初始化策略执行的上下文
        self.context = StrategyContext(kwargs=self.params, timer_factory=self.timer_factory)

        # Analyzer dict, key=name, value=analyzer
        self.analyzers = {}

        # 创建文件日志
        file_path = f"{self.params.get('save_path', 'storage')}/logs/"
        fu.create_dir(path=file_path)
        file_name = f'run_backtest_'
        strat_run_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        file_name += "__time=" + strat_run_time.replace("-", "").replace(":", "").replace(" ", "_") + ".log"
        log_path = file_path + file_name
        # 配置log输出到文件
        log.add_file_handler(file_name=log_path, level=log.DEBUG)

    def add_analyzer(self, name: str, analyzer: Callable):
        self.analyzers[name] = analyzer().set_context(self.context)

    def initialize(self, context):
        """
        初始化方法，在整个回测、模拟实盘中最开始执行一次，用于初始一些全局变量，全局变量会被持久化。重启策略不会再次执行。
        :param context:
        :return:
        """
        if hasattr(self.strategy, 'initialize'):
            self.strategy.initialize(context=context)
        else:
            raise Exception(f"{context.current_dt} 策略中缺少initialize()初始化函数")

    def process_initialize(self, context):
        """
        每次启动策略都会执行的初始化函数，一般用来初始化一些不能持久化保存的内容. , 比如以__开头的全局变量属性，或者计划任务，在 initialize 后执行.
        :param context:
        :return:
        """
        if hasattr(self.strategy, 'process_initialize'):
            self.strategy.process_initialize(context=context)

    def before_trading_start(self, context):
        """
        开盘前运行(可选)
        该函数会在每天开始交易前被调用一次, 可以在这里添加一些每天都要初始化的动作。
        该函数依据的时间是股票的交易时间，即该函数启动时间为'09:00'.
        :param context:
        :return:
        """
        if hasattr(self.strategy, 'before_trading_start'):
            self.strategy.before_trading_start(context=context)

    def after_trading_end(self, context):
        """
        收盘后运行(可选)
        每天结束交易后被调用一次, 您可以在这里添加一些每天收盘后要执行的内容
        :param context:
        :return:
        """
        if hasattr(self.strategy, 'after_trading_end'):
            self.strategy.after_trading_end(context=context)

    def handle_bar(self, context):
        """
        K线处理函数
        :param context:
        :return:
        """
        if hasattr(self.strategy, 'handle_bar'):
            self.strategy.handle_bar(context=context)

    def on_strategy_end(self, context):
        """
        策略结束后执行
        :param context:
        :return:
        """
        if hasattr(self.strategy, 'on_strategy_end'):
            self.strategy.on_strategy_end(context=context)

    @classmethod
    def on_reset(cls, context):
        """
        每日重置的callback函数，必须在before_trading_start()之前执行, 在实盘中，每日早上9:10重置
        :return:
        """
        # --------------------------------------------------------------
        # 重置当日的股票数据字典，包括当日最新价、涨停价、跌停价、是否停牌、是否ST等
        # --------------------------------------------------------------
        log.debug(f"{context.current_dt} 策略每日重置完成")

    @classmethod
    def generate_time_range(cls, start_date, end_date, unit, datetime_format='%Y-%m-%d'):
        """
        获取回测的播放时间范围
        :param start_date:
        :param end_date:
        :param unit:
        :param datetime_format:
        :return:
        """
        # 将字符串日期时间转换为datetime对象
        start_datetime = datetime.datetime.strptime(start_date, datetime_format)
        end_datetime = datetime.datetime.strptime(end_date, datetime_format)
        
        # 设置开始日期时间为开盘时间9:30:00
        start_datetime = start_datetime.replace(hour=9, minute=30, second=0)
        # 设置结束日期时间为收盘时间15:00:00
        end_datetime = end_datetime.replace(hour=15, minute=0, second=0)

        # 初始化时间列表
        time_list = []

        # 确定时间间隔
        if unit == '1d':
            interval = datetime.timedelta(days=1)
        elif unit == '1m':
            interval = datetime.timedelta(minutes=1)
        elif unit == '5m':
            interval = datetime.timedelta(minutes=5)
        elif unit == '15m':
            interval = datetime.timedelta(minutes=15)
        elif unit == '30m':
            interval = datetime.timedelta(minutes=30)
        elif unit == '1h':
            interval = datetime.timedelta(hours=1)
        else:
            raise ValueError(f"Invalid unit: {unit}. Must be '1d', '1m', '5m', '15m', '30m', or '1h'.")

        # 使用循环来递增时间，直到超过结束日期时间
        current_datetime = start_datetime
        while current_datetime <= end_datetime:
            # 对于分钟级别的unit，需要处理交易时间和跨天
            if unit in ['1m', '5m', '15m', '30m', '1h']:
                time_str = current_datetime.strftime('%H%M%S')
                
                # 如果时间在11:30之后、13:00之前，直接跳到13:00（中午休市）
                if '113000' < time_str < '130000':
                    current_datetime = current_datetime.replace(hour=13, minute=0, second=0)
                    if current_datetime > end_datetime:
                        break
                    time_str = current_datetime.strftime('%H%M%S')
                
                # 检查是否在交易时间内
                if not ('093000' <= time_str <= '113000' or '130000' <= time_str <= '150000'):
                    # 不在交易时间内，判断是跨天还是其他情况
                    if time_str > '150000' or time_str < '093000':
                        # 跨天了，跳到下一天的9:30
                        next_day = current_datetime.date() + datetime.timedelta(days=1)
                        current_datetime = datetime.datetime.combine(next_day, datetime.time(9, 30, 0))
                        if current_datetime > end_datetime:
                            break
                        continue
                    else:
                        # 其他情况，继续递增
                        current_datetime += interval
                        continue

            time_list.append(current_datetime)
            current_datetime += interval
            
            # 对于分钟级别的unit，检查递增后的时间，如果跨天了，跳到下一天的9:30
            if unit in ['1m', '5m', '15m', '30m', '1h']:
                time_str = current_datetime.strftime('%H%M%S')
                # 如果时间不在交易时间内（可能是跨天了）
                if not ('093000' <= time_str <= '113000' or '130000' <= time_str <= '150000'):
                    # 判断是跨天还是中午休市
                    if time_str > '150000' or time_str < '093000':
                        # 跨天了，跳到下一天的9:30
                        next_day = current_datetime.date() + datetime.timedelta(days=1)
                        current_datetime = datetime.datetime.combine(next_day, datetime.time(9, 30, 0))
                        if current_datetime > end_datetime:
                            break
                    elif '113000' < time_str < '130000':
                        # 中午休市，跳到13:00
                        current_datetime = current_datetime.replace(hour=13, minute=0, second=0)
                        if current_datetime > end_datetime:
                            break

        return time_list

    def run(self):
        """
        执行回测
        :return:
        """
        start_time = time.time()

        # 执行结果
        results = {"benchmark": self.params['benchmark']}

        # 获取行情数据范围
        log.info(f"回测时间范围：start_date={self.params['start_date']}, end_date={self.params['end_date']}")

        # 获取Benchmark Kline: 最新价格，昨日收盘价，用于计算benchmark的收益率
        # df = api.get_attribute_history(security=self.params['benchmark'],
        #                                fields=['close', 'pre_close'],
        #                                unit=self.params['unit'],
        #                                start_date=self.params['start_date'],
        #                                end_date=self.params['end_date'],
        #                                dividend_type=self.params['dividend_type'])

        # TODO 以get_factor_v2替换get_attribute_history
        df = api.get_factor_v2(symbol=self.params['benchmark'],
                               factor=['close', 'pre_close'],
                               timeframe=self.params['unit'],
                               start_date=self.params['start_date'],
                               end_date=self.params['end_date'],
                               dividend_type=self.params['dividend_type'])

        # 通过层级名称删除symbol
        df = df.droplevel('symbol')

        # 计算收益率
        df['return_rate'] = df['close']/df['pre_close'] - 1

        if self.params['unit'] in ['1d']:
            df.index = pd.to_datetime(df.index)

        # 回测时间范围
        time_range = self.generate_time_range(start_date=self.params['start_date'],
                                              end_date=self.params['end_date'],
                                              unit=self.params['unit'])

        # 按照回测的时间序列向后补充数据，如果有缺失的话
        df = df.reindex(time_range, method='ffill')

        # 设置context.dt为默认的第一天，因为initialize()或者process_initialize()可能会用到
        self.context.set_dt(time_range[0])

        # 初始化现金: 初始化出入金在第一天为回测的初始化资金
        self.context.inout_cash(self.params.get('init_investment', 0.0))

        # 添加默认的analyzer
        self.add_analyzer(name="time_return", analyzer=TimeReturn)

        # 初始化策略, 仅第一次启动策略时执行
        self.initialize(context=self.context)

        # 每次启动策略时执行
        self.process_initialize(context=self.context)

        # 执行Analyzers.start()
        for name, analyzer in self.analyzers.items():
            analyzer.start()

        # 按照回测时间范围的交易日播放进行回测
        trade_date = None  # 当前的日期
        is_open_trade_date = False  # 是否为交易日
        total = len(time_range)
        run_start_dt = datetime.datetime.now()  # 用于估算回测还剩余多少时间
        # 使用 enumerate() 和 tqdm()
        for index, trade_time in enumerate(tqdm(time_range, total=len(time_range))):
            # 在这里你可以使用 index 和 trade_time
            # 注意：由于 tqdm 是基于迭代器工作的，它实际上并不需要 total 参数来正确工作，
            # 但提供 total 可以让进度条更准确地显示。
            # 如果不知道长度，可以省略 total 参数，但进度条可能不会非常准确。
            # 这里的 trade_time 实际上是枚举的值（即 range 中的数字）

            # task打点进度
            if self.task is not None:
                run_current_dt = datetime.datetime.now()
                run_delta = (run_current_dt - run_start_dt).seconds
                self.task.update_state(meta={"current": index + 1,
                                             "estimate_left_time": int(run_delta/(index + 1) * (total - index - 1)),
                                             "total": total})

            # new trade date, 每天执行一次after_trading_end()，最后一次在for循环结束后执行
            if trade_date is not None and trade_time.strftime('%Y-%m-%d') != trade_date:
                # 每日收盘后执行，必须在更新设置context之前执行
                self.after_trading_end(context=self.context)

            # new trade date，每天执行一次，检查是否为交易日
            if trade_date is None or trade_time.strftime('%Y-%m-%d') != trade_date:
                is_open_trade_date = api.is_open_trade_date(trade_time.strftime('%Y-%m-%d'))

            # 记录benchmark的收盘价，每个bar执行一次, 在unit=‘1d’，需要设置在context.set_dt()之前
            if is_open_trade_date:
                # 设置benchmark的收益率
                self.context.set_benchmark_value(df.loc[trade_time]['return_rate'])

            # 设置context.dt，每个bar执行一次，必须在on_reset()和before_trading_start()之前，因context.dt会在这些函数内被访问
            if self.context.unit in ['1d']:
                # 如果行情周期unit为天，则回测的当前时间(context.current_dt)指向开票时间9:30
                open_09_30 = datetime.datetime(trade_time.year, trade_time.month, trade_time.day, 9, 30, 0)
                self.context.set_dt(open_09_30)
            else:
                self.context.set_dt(trade_time)

            # new trade date，每天执行一次，每日重置on_reset，看盘前运行before_trading_start的函数
            if trade_date is None or trade_time.strftime('%Y-%m-%d') != trade_date:
                trade_date = trade_time.strftime('%Y-%m-%d')
                # 每日重置
                if is_open_trade_date:
                    self.on_reset(context=self.context)

                # 每日开盘前，必须执行在定时器任务之前
                if is_open_trade_date:
                    self.before_trading_start(context=self.context)

            # 如果主图按照1d的数据回测，默认在收盘时执行当天的任务
            if self.context.unit in ['1d']:
                # 如果行情周期unit为天，则回测的当前时间(context.current_dt)指向收盘时间15:00
                close_15_00 = datetime.datetime(trade_time.year, trade_time.month, trade_time.day, 15, 0, 0)
                self.context.set_dt(close_15_00)

            # 执行定时器任务:run_daily, run_weekly, run_monthly, run_periodically
            self.timer_factory.notify_timer(self.context.current_dt)

            # 按K线图执行
            if is_open_trade_date:
                self.handle_bar(context=self.context)

            # 执行Analyzers.next()
            if is_open_trade_date:
                for name, analyzer in self.analyzers.items():
                    analyzer.next()
        else:
            # 最后一个Bar收盘后执行
            if is_open_trade_date:
                self.after_trading_end(context=self.context)

        # 策略结束
        self.on_strategy_end(context=self.context)

        # 执行Analyzers.stop()
        for name, analyzer in self.analyzers.items():
            analyzer.stop()

        # 获取Analyzers结果
        analysis = {}
        for name, analyzer in self.analyzers.items():
            analysis[name] = analyzer.get_analysis()

        results['analysis'] = analysis

        # 画图数据
        results['plot_data'] = self.context.plot_data

        # 所有委托明细
        orders_df = pd.DataFrame(self.context.orders)
        results['orders'] = orders_df

        # 所有出入金历史
        results['inout_cash'] = pd.DataFrame(self.context.inout_cash_his)

        # 总回测时长
        end_time = time.time()
        # 返回累计收益结果
        strat_returns = results['analysis']['time_return']['strat_return']
        bchmk_returns = results['analysis']['time_return']['bchmk_return']
        cum_strat_returns = (1 + strat_returns).cumprod() - 1
        cum_bchmk_returns = (1 + bchmk_returns).cumprod() - 1
        log.info(f"时长(m)={(end_time - start_time)/60:.2f}, "
                 f"策略收益={cum_strat_returns.values.tolist()[-1]} "
                 f"基准收益={cum_bchmk_returns.values.tolist()[-1]} "
                 )

        return results


# ---------------------------------------------------------
# 外部可以访问的列表
# ---------------------------------------------------------
__all__ = ["BacktestExecutor"]
__all__.extend([name for name in globals().keys() if name.startswith("get")])



from six import with_metaclass
import abc
from pqsdk.enums.orderStatus import OrderStatus
import datetime


class AbstractOrder(with_metaclass(abc.ABCMeta)):
    """
    买卖订单
    """

    @property
    @abc.abstractmethod
    def status(self) -> OrderStatus:
        """
        状态, 一个OrderStatus值
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def add_time(self) -> datetime.datetime:
        """
        订单添加时间, [datetime.datetime]对象
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def is_buy(self) -> bool:
        """
        bool值, 买还是卖。对于期货: (1) 开多/平空 -> 买; （2）开空/平多 -> 卖
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def volume(self):
        """
        下单数量, 不管是买还是卖, 都是正数
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def filled(self):
        """
        已经成交的股票数量, 正数
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def security(self):
        """
        股票代码
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def order_id(self):
        """
        订单ID
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def price(self):
        """
        平均成交价格, 已经成交的股票的平均成交价格(一个订单可能分多次成交)
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def avg_cost(self):
        """
        卖出时表示下卖单前的此股票的持仓成本, 用来计算此次卖出的收益. 买入时表示此次买入的均价(等同于price).
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def commission(self):
        """
        交易费用（佣金、税费等）
        :return:
        """
        raise NotImplementedError



import logging
import colorlog
import os
import threading
from logging.handlers import RotatingFileHandler


class Logger(logging.Logger):
    # 日志级别：（critical > error > warning > info > debug）
    CRITICAL = logging.CRITICAL
    FATAL = logging.FATAL
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    WARN = logging.WARN
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    NOTSET = logging.NOTSET

    # 单例模式实现
    _instances = {}
    _lock = threading.Lock()

    @classmethod
    def get_instance(cls, name=None, **kwargs):
        """
        获取Logger实例，如果已存在则返回现有实例，否则创建新实例
        :param name: logger名称
        :param kwargs: 其他参数
        :return: Logger实例
        """
        if name is None:
            name = 'default_pq_logger'

        with cls._lock:
            if name not in cls._instances:
                cls._instances[name] = cls(name, **kwargs)
            return cls._instances[name]

    def __init__(self, name=None, file_name=None, console_level=None, file_level=None,
                 max_bytes=10 * 1024 * 1024, backup_count=5, date_format='%Y-%m-%d %H:%M:%S'):
        """
        自定义日志对象
        :param name: logger名称，默认输出日志到控制台
        :param file_name: 输出文件路径，为None则不输出日志到文件
        :param console_level: 控制台日志级别
        :param file_level: 日志文件的日志级别
        :param max_bytes: 单个日志文件的最大大小，默认10MB
        :param backup_count: 保留的日志文件数量，默认5个
        :param date_format: 日期格式，默认'%Y-%m-%d %H:%M:%S'
        """
        if name is None:
            name = 'default_pq_logger'

        # 调用父类初始化
        super().__init__(name)

        # 阻止日志消息从当前Logger向父级Logger传递
        self.propagate = False

        if console_level is None:
            console_level = self.DEBUG

        if file_level is None:
            file_level = self.WARNING  # 默认只有warning及以上级别才会写入日志文件

        # 指定最低日志级别：（critical > error > warning > info > debug）
        self.setLevel(level=self.DEBUG)

        # 控制台输出不同级别日志颜色设置
        self.color_config = {
            'DEBUG': 'cyan',
            'INFO': 'green',
            'WARNING': 'yellow',
            'ERROR': 'red',
            'CRITICAL': 'purple',
        }

        # 保存配置
        self.date_format = date_format
        self.max_bytes = max_bytes
        self.backup_count = backup_count

        # -------------------------
        # 输出到控制台
        # 日志格化字符串
        # -------------------------
        self.add_console_handler(level=console_level)

        # -------------------------
        # 输出到文件
        # -------------------------
        if file_name:
            self.add_file_handler(file_name=file_name, level=file_level)

    def __enter__(self):
        """
        上下文管理器入口
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        上下文管理器出口，关闭所有处理器
        """
        self.close()

    def close(self):
        """
        关闭所有处理器
        """
        for handler in self.handlers[:]:
            try:
                handler.close()
                self.removeHandler(handler)
            except Exception as e:
                # 记录错误但不抛出异常
                print(f"关闭处理器时出错: {e}")

    def set_name(self, name):
        """
        设置logger的名称，如果此名称的logger不存在，则创建一个新的logger
        如果新建的logger，需要通过：add_console_handler()或者add_file_handler()输出到控制台或者文件
        :param name: logger名称
        :return:
        """
        # 由于Logger类继承自logging.Logger，这里需要重新创建一个新的Logger实例
        # 保存当前的handlers
        handlers = self.handlers[:]
        # 清除当前handlers
        for handler in handlers:
            self.removeHandler(handler)

        # 创建新的Logger实例
        new_logger = Logger(name=name)

        # 将当前实例的属性复制到新实例
        self.__dict__.update(new_logger.__dict__)

        # 重新添加handlers
        for handler in handlers:
            self.addHandler(handler)

    def set_level(self, level=logging.INFO):
        """
        设置日志级别
        :param level: 日志级别
        """
        self.setLevel(level)

    def add_console_handler(self, level=logging.DEBUG, fmt=None):
        """
        输出到控制台
        :param level: 日志级别
        :param fmt: 日志格式，如果为None则使用默认格式
        :return:
        """
        try:
            if fmt is None:
                fmt = '%(log_color)s%(asctime)s: %(levelname)s %(message)s'

            console_formatter = colorlog.ColoredFormatter(
                fmt=fmt,
                log_colors=self.color_config,
                datefmt=self.date_format
            )
            console_handler = logging.StreamHandler()
            console_handler.setFormatter(console_formatter)
            console_handler.setLevel(level)
            self.addHandler(console_handler)
        except Exception as e:
            print(f"添加控制台处理器时出错: {e}")

    def add_file_handler(self,
                         file_name: str,
                         level=logging.WARNING,
                         fmt=None,
                         max_bytes=None,
                         backup_count=None,
                         json: bool = False,
                         verbose: bool = False):
        """
        输出到文件
        :param file_name: 日志文件路径
        :param level: 日志级别
        :param fmt: 日志格式，如果为None则使用默认格式
        :param max_bytes: 单个日志文件的最大大小，如果为None则使用默认值
        :param backup_count: 保留的日志文件数量，如果为None则使用默认值
        :param json: 是否以JSON格式输出日志，默认为False
        :param verbose: 是否使用详细的JSON格式（仅当json=True时有效），默认为False
        :return:
        """
        try:
            # 检查是否已经存在相同文件名的FileHandler
            file_path = os.path.abspath(file_name)
            for handler in self.handlers:
                if isinstance(handler,
                              (logging.FileHandler, RotatingFileHandler)) and handler.baseFilename == file_path:
                    # 如果已存在相同文件名的FileHandler，则更新其日志级别
                    handler.setLevel(level)
                    return

            # 如果不存在相同文件名的FileHandler，则创建新的
            import json_log_formatter

            # 确保日志目录存在
            log_dir = os.path.dirname(file_path)
            if log_dir and not os.path.exists(log_dir):
                os.makedirs(log_dir)

            # 使用RotatingFileHandler替代FileHandler，支持日志轮转
            max_bytes = max_bytes or self.max_bytes
            backup_count = backup_count or self.backup_count

            file_handler = RotatingFileHandler(
                filename=file_name,
                mode='a',
                encoding='utf-8',
                maxBytes=max_bytes,
                backupCount=backup_count
            )

            if fmt is None:
                if json:
                    if verbose:
                        file_formatter = json_log_formatter.VerboseJSONFormatter()
                    else:
                        file_formatter = json_log_formatter.JSONFormatter()
                else:
                    file_formatter = logging.Formatter(
                        fmt='%(asctime)s: %(levelname)s %(message)s',
                        datefmt=self.date_format
                    )
            else:
                file_formatter = logging.Formatter(fmt=fmt, datefmt=self.date_format)

            file_handler.setFormatter(file_formatter)
            file_handler.setLevel(level)
            self.addHandler(file_handler)
        except Exception as e:
            print(f"添加文件处理器时出错: {e}")

    def remove_file_handler(self, file_name: str):
        """
        删除输出到文件的handler
        :param file_name: 日志文件路径
        :return:
        """
        try:
            # 遍历logger的handlers列表
            for handler in list(self.handlers):
                # 检查handler是否是FileHandler或RotatingFileHandler且其文件名匹配给定的文件名
                if isinstance(handler, (logging.FileHandler, RotatingFileHandler)):
                    # handler.baseFilename 为绝对路径
                    if handler.baseFilename == os.path.abspath(file_name):
                        # 从logger的handlers列表中移除handler
                        self.removeHandler(handler)
                        # 关闭handler以释放资源
                        handler.close()
                        # 删除对它的引用
                        del handler
                        break  # 如果只期望移除一个匹配的handler，则退出循环
        except Exception as e:
            print(f"移除文件处理器时出错: {e}")


# 创建默认的全局logger
log = Logger.get_instance(name='default_pq_logger')

# ---------------------------------------------------------
# 外部可以访问的列表
# ---------------------------------------------------------
__all__ = ["log", "Logger"]
__all__.extend([name for name in globals().keys() if name.startswith("get")])



from six import with_metaclass
import abc
from .abstractOrder import AbstractOrder


class AbstractTrader(with_metaclass(abc.ABCMeta)):

    @abc.abstractmethod
    def buy(self, stock_code: str, volume: int, price_type: int, price: float, force: bool = False) -> AbstractOrder:
        """
        证券买入委托
        Returns:
          - 委托的order_id
        """
        raise NotImplementedError

    @abc.abstractmethod
    def sell(self, stock_code: str, volume: int, price_type: int, price: float, force: bool = False) -> AbstractOrder:
        """
        证券卖出委托
        Returns:
          - 委托的order_id
        """
        raise NotImplementedError

    @abc.abstractmethod
    def order_target_percent(self, stock_code: str, target: float = 0.0, force: bool = False) -> AbstractOrder:
        """
        基于order_target_value（），按照总资产的百分比委托下单

        :param force: 是否强制交易
        :param stock_code:目标持仓股票
        :param target: 目标持仓比例
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def order_target_value(self, stock_code: str, target: float = 0.0, price: float = None,
                           force: bool = False) -> AbstractOrder:
        """
        Place an order to rebalance a position to have final value of
        ``target``

        The current ``value`` is taken into account as the start point to
        achieve ``target``

          - If no ``target`` then close postion on data
          - If ``target`` > ``value`` then buy on data
          - If ``target`` < ``value`` then sell on data

        It returns either:

          - The generated order

          or

          - ``None`` if no order has been issued

        :param force: 是否强制交易
        :param stock_code: 目标持仓股票
        :param target: 目标持仓市值
        :param price:
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def order_target_volume(self, stock_code: str, target: int = 0.0, force: bool = False) -> AbstractOrder:
        """
        基于order_target_value()，按照个股的数量目标数量下单。注意，可能存在账户资金不足的情况。

        :param force: 是否强制交易
        :param stock_code: 目标持仓股票
        :param target: 目标持仓数量
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def close(self, stock_code: str, force: bool = False) -> AbstractOrder:
        """
        以最新价平仓股票代码, 默认平仓数量为可用数量
        :param force: 是否强制交易
        :param stock_code: 需要平仓的股票代码
        :return:
        """
        raise NotImplementedError



from pqsdk.interface import AbstractRunInfo
from pqsdk.api import get_index_members_lst, get_stock_pool


class RunInfo(AbstractRunInfo):

    def __init__(self, kwargs: dict, context):
        self.__kwargs = kwargs
        self.context = context

    @property
    def tenant_id(self):
        return self.__kwargs.get('user').tenant_id

    @property
    def portfolio_id(self):
        """
        回测的时候，投资组合可能还未指定
        :return:
        """
        return -1

    @property
    def parameters(self):
        return self.__kwargs.get('parameters', {})

    @property
    def start_date(self):
        return self.__kwargs.get('start_date')

    @property
    def end_date(self):
        return self.__kwargs.get('end_date')

    @property
    def stock_starting_cash(self):
        return self.__kwargs.get('init_investment')

    @property
    def strategy_id(self):
        return self.__kwargs.get('strategy_id')

    @property
    def release_id(self):
        """
        回测的时候，策略代码可能还未发版
        :return:
        """
        return -1

    @property
    def strategy_name(self):
        return self.__kwargs.get('strategy_name')

    @property
    def strategy_remark(self):
        """
        策略备注
        :return:
        """
        return 'backtest'

    @property
    def is_queued_order(self):
        """
        回测中不拦截委托
        :return:
        """
        return False

    @property
    def benchmark(self):
        return self.__kwargs.get('benchmark')

    @property
    def index(self):
        return self.__kwargs.get('index', [])

    @property
    def stock_pool(self) -> list:
        return self.__kwargs.get('stock_pool', [])

    @property
    def stock_pool_members(self) -> list:
        """
        当日的股票池成员, 如果用户选择了指数成分股参数，则在股票池中过滤指数成分股
        :return:
        """
        index_members = []
        if len(self.index) > 0:
            index_members = get_index_members_lst(index_lst=self.index,
                                                  trade_date=self.context.current_dt.strftime('%Y-%m-%d'))

        stock_pool_members = []
        if len(self.stock_pool) > 0:
            stock_pool_members = get_stock_pool(ids=self.stock_pool)  # 通过pqsdk获取股票池成员
            if len(index_members) > 0:
                stock_pool_members = [stock for stock in stock_pool_members if stock in index_members]
        return stock_pool_members if len(stock_pool_members) > 0 else index_members



from pqsdk.interface import AbstractPortfolio
from pqsdk.api import get_factor_v2


class Portfolio(AbstractPortfolio):

    def __init__(self, context):
        """
        策略的投资组合

        """

        self._positions = {}
        self._cash = 0.0
        self.context = context

    @property
    def positions(self) -> dict:
        """
        [dict] 持仓字典
        {
        "510300.SH": Position(stock_code, Position)}
        其中，Position包含如下属性：
        {
          "volume": 2600,
          "can_use_volume": 1400,
          "open_price": 20.206,
          "market_value": 10701.599999999999
        }
        """
        # # 清除已经被平仓的头寸
        # for sec_code, position in self._positions.items():
        #     if position.volume == 0:
        #         del self._positions[sec_code]

        return self._positions

    @property
    def total_value(self):
        return self.positions_value + self.available_cash

    @property
    def positions_value(self):
        # TODO 获取当时的持仓价值
        if self.context.unit in ['1d']:
            trade_date = self.context.current_dt.strftime('%Y-%m-%d')
            trade_datetime = None
        else:
            trade_date = None
            trade_datetime = self.context.current_dt.strftime('%Y-%m-%d %H:%M:%S')

        sec_code_lst = list(self.positions.keys())
        if len(sec_code_lst) == 0:
            # 策略无持仓
            return 0
        # count=1, 意思是：截止trade_date的最新价格，如果trade_date停牌，将提取最近的价格
        # price_df = get_history(count=1,
        #                        end_date=trade_date,
        #                        end_datetime=trade_datetime,
        #                        unit=self.context.unit,
        #                        field='close',
        #                        security_list=sec_code_lst,
        #                        dividend_type=self.context.dividend_type,
        #                        expect_df=True)

        # 使用get_factor_v2获取因子值
        # limit=1, 意思是：截止trade_date的最新价格，如果trade_date停牌，将提取最近的价格
        price_df = get_factor_v2(limit=1,
                                 end_date=trade_date,
                                 end_datetime=trade_datetime,
                                 timeframe=self.context.unit,
                                 factor='close',
                                 symbol=sec_code_lst,
                                 dividend_type=self.context.dividend_type,
                                 expect_df=True)

        # 或通过层级位置删除（symbol是第0层, trade_date或者datetime是第1层）
        price_df = price_df.droplevel(1)

        price_dict = price_df.iloc[:, 0].to_dict()

        total_value = 0.0
        for sec_code, position in self.positions.items():
            price = price_dict.get(sec_code, 0.0)
            volume = position.volume
            total_value += volume * price

        return total_value

    @property
    def available_cash(self):
        return self._cash

    def add_cash(self, cash: float):
        """
        添加/减少可用金额
        :param cash:
        :return:
        """
        self._cash = self._cash + cash

    def get_frozen_cash(self, sec_code: str = None):
        """
        获取冻结现金
        冻结现金 = sum（（买入委托数量 - 买入成交数量） * 委托价格）

        在回测过程中，买入委托即刻成交，所以冻结现金默认为0

        :param sec_code: 获取指定股票的冻结现金, 如果为None，则获取投资组合的冻结现金
        :return:
        """
        return 0



from .abstractStrategyContext import AbstractStrategyContext
from .abstractInstrument import AbstractInstrument
from .abstractPosition import AbstractPosition
from .constant import Constant
from .abstractOrder import AbstractOrder
from .abstractPortfolio import AbstractPortfolio
from .abstractRunInfo import AbstractRunInfo
from .abstractTrader import AbstractTrader

__all__ = [
    "AbstractStrategyContext",
    "AbstractInstrument",
    "AbstractPosition",
    "Constant",
    "AbstractOrder",
    "AbstractPortfolio",
    "AbstractRunInfo",
    "AbstractTrader"
]



# coding=utf-8
import pqsdk.backtest as backtest
import pqsdk.utils.file_util as fu
from .logger import log


def run_backtest(file_path: str,
                 parameters: dict,
                 plot: bool,
                 save_metrics: bool,
                 save_orders: bool,
                 save_inout_cash: bool
                 ):
    """
    执行策略回测

    :param file_path: 是回测脚本的路径
    :param parameters: 是一个字典，包含了所有通过 -p 传递的参数
    :param plot: 是一个布尔值，表示是否绘图
    :param save_metrics: 是否保存回测结果
    :param save_orders: 是否保存委托明细信息
    :param save_inout_cash: 是否保存出入金历史
    :return:
    """

    # 回测参数
    params = {
        "cash": parameters.get('cash', 1000000),  # 初始资金
        "start_date": parameters.get("start_date", "2024-03-01"),  # 回测开始日期
        "end_date": parameters.get("end_date", "2024-03-17"),  # 回测介绍日期
        "benchmark": parameters.get("benchmark", "000300.SH"),  # 行情基准
        "stock_pool": parameters.get("stock_pool", "000300.SH").split(","),  # 逗号分隔的股票池
        "unit": parameters.get("unit", '1d'),  # 行情周期
        "adjust_period": parameters.get("adjust_period", 5),  # 调仓周期，默认5个交易日
        "hold_maxsize": parameters.get("hold_maxsize", 10),  # 默认选股个数
    }

    # 添加额外的参数
    for k, v in parameters.items():
        if k not in params:
            params[k] = v

    # 策略代码
    if not fu.check_path_exists(file_path):
        log.error(f"输入的策略文件路径不存在: -f {file_path}")
        exit(-1)
    with open(file_path, 'r', encoding='utf-8') as f:
        script = f.read()

    # 执行回测
    results = backtest.execute(parameters=params, script=script)

    if plot:
        backtest.tearsheet(results)

    if save_metrics:
        backtest.save_metrics(results=results)

    if save_orders:
        backtest.save_orders(results=results)

    if save_inout_cash:
        backtest.save_inout_cash(results=results)





import datetime
from abc import ABC
from typing import Callable, Union, List
import pandas as pd
from pqsdk.interface import AbstractStrategyContext, AbstractInstrument, AbstractPosition, Constant
from pqsdk.utils.timer_factory import TimerFactory
from .portfolio import Portfolio
from .position import Position
from .order import Order
from pqsdk import log, pqconstant
from pqsdk.api import get_previous_trading_date, get_factor_v2, get_stock_info_v2
from .run_info import RunInfo


class StrategyContext(AbstractStrategyContext, ABC):

    def __init__(self, kwargs: dict, timer_factory: TimerFactory):
        # 自定义定时器
        self.__tf = timer_factory
        self.__constant = Constant()  # 初始化常量
        # 策略执行信息
        self.__run_info = RunInfo(kwargs=kwargs, context=self)
        # 回测的当前日期，每日set_dt重置
        self.__dt = None
        # 回测当前日期的Benchmark收益率
        self.__benchmark_value = None
        self.__unit = kwargs.get('unit', '1d')  # 行情周期
        self.__parameters = kwargs.get('parameters')
        self.__dividend_type = kwargs.get('dividend_type', 'back')
        self.__strategy_name = kwargs.get('strategy_name', None)
        self.__commission = kwargs.get('commission', 0.0)  # 佣金费率
        self.__index = kwargs.get('index', [])
        self.__stock_pool = kwargs.get('stock_pool', [])

        # 初始化Portfolio
        self._portfolio = Portfolio(self)

        # 初始化委托编号
        self.order_id = 0

        # 历史委托列表, 从notify_order()函数收集委托数据，回测结束后写入数据库
        self.orders = []

        """
        回测过程中画出曲线
        self.plot_data = {"000001.SZ": {"startExitPrice": {"type": "line", "xAxis": [], "yAxis": [], "desc": ""}}}
        """
        self.plot_data = dict()

        """
        记录投资组合的转入或转出资金历史
        self.inout_cash_his = [{"datetime": self.current_dt: "cash": cash}]
        """
        self.inout_cash_his = []

    def notify_order(self, order):
        """
        新创建委托的通知
        :param order: 委托的对象
        :return:
        """
        order_dict = {"order_id": order.order_id,
                      "sec_code": order.security,
                      "volume": order.volume,
                      "price": order.price,
                      "is_buy": order.is_buy,
                      "avg_cost": order.avg_cost,
                      "comm": order.commission,
                      "add_time": order.add_time,
                      "trade_date": order.add_time.strftime('%Y-%m-%d')}
        self.orders.append(order_dict)
        # log.debug(f"order: {order_dict}")

    def plot_line(self, sec_code: str, name: str, value: float, desc=None):
        """
        画曲线
        :param sec_code:
        :param name:
        :param value:
        :param desc:
        :return:
        """
        self.__plot_chart(chart_type='line', sec_code=sec_code, name=name, value=value, desc=desc)

    def plot_bar(self, sec_code: str, name: str, value: float, desc=None):
        """
        画柱状图
        :param sec_code:
        :param name:
        :param value:
        :param desc:
        :return:
        """
        self.__plot_chart(chart_type='bar', sec_code=sec_code, name=name, value=value, desc=desc)

    def __plot_chart(self, chart_type: str, sec_code: str, name: str, value: float, desc=None):
        """
        画图表
        :param chart_type: chart类型：line， bar
        :param sec_code: 股票代码
        :param name: 曲线名称
        :param value: 曲线yAxis的值
        :param desc: value的描述
        :return:
        """
        if self.unit in ['1d']:
            dt = self.current_dt.strftime('%Y-%m-%d')
        else:
            dt = self.current_dt.strftime('%Y-%m-%d %H:%M')
        if sec_code not in self.plot_data:
            self.plot_data[sec_code] = {name: {"type": chart_type,
                                               "xAxis": [dt],
                                               "yAxis": [round(value, 2)],
                                               "desc": [desc]}}
        elif name not in self.plot_data[sec_code]:
            self.plot_data[sec_code][name] = {"type": chart_type,
                                              "xAxis": [dt],
                                              "yAxis": [round(value, 2)],
                                              "desc": [desc]}
        else:
            self.plot_data[sec_code][name]["xAxis"].append(dt)
            self.plot_data[sec_code][name]["yAxis"].append(round(value, 2))
            self.plot_data[sec_code][name]["desc"].append(desc)

    def run_monthly(self, func: Callable, monthday: int, time_str: str) -> None:
        """
        按月运行

        :param func: 一个自定义的函数, 此函数必须接受context参数;例如自定义函数名market_open(context)
        :param monthday: 每月的第几个交易日
        :param time_str: 具体执行时间,一个字符串格式的时间, 24小时内的任意时间，例如"10:00", "01:00"；在实盘中才生效
        """

        self.__tf.add_timer(callback=func, kwargs={"context": self}, when=time_str, monthdays=[monthday],
                            monthcarry=True)

    def run_weekly(self, func: Callable, weekday: int, time_str: str) -> None:
        """
        按周运行

        :param func: 一个自定义的函数, 此函数必须接受context参数;例如自定义函数名market_open(context)
        :param weekday: 每周的第几个交易日, 1 = monday, ... 7 = sunday
        :param time_str: 具体执行时间,一个字符串格式的时间, 24小时内的任意时间，例如"10:00", "01:00"；在实盘中才生效
        """
        self.__tf.add_timer(callback=func, kwargs={"context": self}, when=time_str, weekdays=[weekday], weekcarry=True)

    def run_periodically(self,
                         func: Callable,
                         days: int,
                         time_str: str,
                         start_date: str = None,
                         end_date: str = None) -> None:
        """
        按调仓周期运行，如果遇到调仓日为非交易日，则顺延到下一个交易日。

        :param func: 一个自定义的函数, 此函数必须接受context参数;例如自定义函数名market_open(context)
        :param days : 间隔天数执行，start_date为第一次执行日期
        :param time_str: 具体执行时间,一个字符串格式的时间, 24小时内的任意时间，例如"10:00", "01:00"；在实盘中才生效
        :param start_date: 开始日期，默认为回测开始日期
        :param end_date: 结束日期, 默认为回测结束日期
        """
        if start_date is None:
            start_date = self.run_info.start_date
        if end_date is None:
            end_date = self.run_info.end_date

        # 将字符串日期转换为datetime对象
        start_date_ = datetime.datetime.strptime(start_date, '%Y-%m-%d')
        end_date_ = datetime.datetime.strptime(end_date, '%Y-%m-%d')

        # 初始化日期列表
        dates = []
        # 当当前日期小于或等于结束日期时，将其添加到列表中，并向前移动interval_days天
        current_date = start_date_
        while current_date <= end_date_:
            dates.append(current_date.strftime('%Y-%m-%d'))
            current_date += datetime.timedelta(days=days)

        self.__tf.add_timer(callback=func,
                            kwargs={"context": self},
                            when=time_str,
                            dates=dates,
                            datecarry=True)

    def run_daily(self, func: Callable, time_str: str) -> None:
        """
        每天内何时运行

        :param func: 一个自定义的函数, 此函数必须接受context参数;例如自定义函数名market_open(context)
        :param time_str: 具体执行时间,一个字符串格式的时间, 24小时内的任意时间，例如"10:00", "01:00"；在实盘中才生效
        """
        self.__tf.add_timer(callback=func, kwargs={"context": self}, when=time_str)

    def run_minutely(self, func: Callable, time_str: str, minutes: int = 1) -> None:
        """
        每天内何时开始，间隔n分钟执行

        :param func: 一个自定义的函数, 此函数必须接受context参数;例如自定义函数名market_open(context)
        :param time_str: 具体执行时间,一个字符串格式的时间, 24小时内的任意时间，例如"10:00", "01:00"；
        :param minutes: 间隔执行的分钟数
        """
        self.__tf.add_timer(callback=func, kwargs={"context": self}, when=time_str, minutes=minutes)

    def run_secondly(self, func: Callable, time_str: str, seconds: int = 1) -> None:
        """
        每天内何时开始，间隔n秒执行

        :param func: 一个自定义的函数, 此函数必须接受context参数;例如自定义函数名market_open(context)
        :param time_str: 具体执行时间,一个字符串格式的时间, 24小时内的任意时间，例如"10:00", "01:00"；
        :param seconds: 间隔执行的秒数
        """
        self.__tf.add_timer(callback=func, kwargs={"context": self}, when=time_str, seconds=seconds)

    def cancel(self, order, force: bool = False):
        pass

    def get_current_data(self):
        """
        获取当前单位时间（当天/当前分钟）的涨跌停价, 是否停牌，当天的开盘价等

        :return: 一个dict, 其中 key 是股票代码, value 是拥有如下属性的对象, 返回的结果只在当天有效:
                last_price : 最新价
                high_limit: 涨停价
                low_limit: 跌停价
                paused: 是否停止或者暂停了交易, 当停牌、未上市或者退市后返回 True
                is_st: 是否是 ST(包括ST, *ST)，是则返回 True，否则返回 False
                day_open: 当天开盘价
                name: 股票现在的名称, 可以用这个来判断股票当天是否是 ST, *ST, 是否快要退市
                industry_code: 股票现在所属行业代码
        """
        return self.current_data_dict

    def get_position(self, stock_code) -> AbstractPosition:
        positions = self.portfolio.positions
        if stock_code in positions:
            return positions[stock_code]
        else:
            # 返回一个空的持仓对象
            return Position(stock_code=stock_code, volume=0, price=0, init_time=datetime.datetime.now(), context=self)

    @property
    def constant(self):
        return self.__constant

    @property
    def commission(self):
        """
        佣金费率
        :return:
        """
        return self.__commission

    @property
    def previous_trade_date(self) -> datetime.date:
        pre_trade_date_str = get_previous_trading_date(self.current_dt.strftime('%Y-%m-%d'))
        return datetime.datetime.strptime(pre_trade_date_str, '%Y-%m-%d').date()

    @property
    def run_info(self):
        return self.__run_info

    def get_orders(self, trade_date: str = None, is_buy: bool = None, include_canceled: bool = True):
        """
        获取交易日的委托单，返回DataFrame

        :param include_canceled: 是否包括已撤的委托
        :param trade_date:
        :param is_buy:
        :return: DataFrame,
            字段：['order_code', 委托编号
                'sec_code',  股票代码
                'order_time', 委托时间
                'is_buy', 委托买卖方向：买单=1， 卖单=0
                'order_volume',  委托数量
                'order_price']  委托价格
        """
        df = pd.DataFrame(self.orders)
        df.rename(columns={"order_id": "order_code",
                           "add_time": "order_time",
                           "volume": "order_volume",
                           "price": "order_price"}, inplace=True)

        if not df.empty and trade_date is not None:
            df = df[df['trade_date'] == trade_date]

        if not df.empty:
            if is_buy is not None:
                df = df[df['is_buy'] == is_buy]

            df['is_buy'] = df['is_buy'].apply(lambda x: 1 if True else 0)

            # 与实盘保持输出一致
            df = df[['order_code', 'sec_code', 'order_time', 'is_buy', 'order_volume', 'order_price']]
        return df

    def set_benchmark_value(self, value: float):
        """
        设置回测过程中当前日期的benchmark的收益率
        :param value:
        :return:
        """
        self.__benchmark_value = value

    @property
    def benchmark_value(self):
        """
        当前日期的benchmark close价格
        :return:
        """
        return self.__benchmark_value

    def set_dt(self, dt):
        self.__dt = dt

    @classmethod
    def __calculate_size(cls, price, cash):
        """
        返回指定价格和现金的股数，需要考虑一手为100股的情况。

        :param price:
        :param cash:
        :return:
        """
        return int(cash // price // 100) * 100

    def order_target_value(self, stock_code: str, target: float = 0.0, price: float = None, force: bool = False):
        # 获取价格
        if price is None:
            # trade_date = self.current_dt.strftime('%Y-%m-%d')
            if self.unit in ['1d']:
                trade_date = self.current_dt.strftime('%Y-%m-%d')
                trade_datetime = None
            else:
                trade_date = None
                trade_datetime = self.current_dt.strftime('%Y-%m-%d %H:%M:%S')
            # price_df = get_history(count=1,
            #                        end_date=trade_date,
            #                        end_datetime=trade_datetime,
            #                        unit=self.unit,
            #                        field='close',
            #                        security_list=[stock_code],
            #                        dividend_type=self.dividend_type,
            #                        expect_df=True)

            # 从get_factor_v2获取
            price_df = get_factor_v2(limit=1,
                                     end_date=trade_date,
                                     end_datetime=trade_datetime,
                                     timeframe=self.unit,
                                     factor='close',
                                     symbol=[stock_code],
                                     dividend_type=self.dividend_type,
                                     expect_df=True)

            if price_df.empty:
                log.warning(f"股票停牌，取消买入: stock_code={stock_code}, trade_date={trade_date}")
                return
            else:
                price = price_df.iloc[0, 0]

        position: Position = self.portfolio.positions.get(stock_code, None)
        if position is None:
            market_value = 0.0
        else:
            market_value = position.volume * price

        if target > market_value:  # Buy
            size = self.__calculate_size(price=price, cash=target - market_value)
            if size < 100:
                content = f"策略名称={self.strategy_name}, 类型=Buy, 股票代码={stock_code}, " \
                          f"目标市值={target}, 持仓市值={market_value}, " \
                          f"委托价格={round(price, 2)},交易股数不足100股，放弃交易"
                log.warning(content)
                return None  # 未满足交易条件
            else:
                return self.buy(stock_code=stock_code,
                                volume=size,
                                price_type=pqconstant.FIX_PRICE,
                                price=price,
                                force=force)

        elif target < market_value:  # Sell
            # 计算可交易股数
            if int(target) == 0:  # 平仓
                size = position.can_use_volume
            else:
                size = self.__calculate_size(price=price, cash=market_value - target)

            if size < 100:
                content = f"策略名称={self.strategy_name}, 类型=Sell, 股票代码={stock_code}, " \
                          f"目标市值={target}, 持仓市值={market_value}, 持仓可用数量={position.can_use_volume}，" \
                          f"委托价格={round(price, 2)},委托数量={size}, 交易股数不足100股，放弃交易"
                log.warning(content)
                return None  # 未满足交易条件
            else:
                if position.can_use_volume >= size >= 100:
                    content = f"策略名称={self.strategy_name}, 类型=Sell, 股票代码={stock_code}, " \
                              f"目标市值={target}, 持仓市值={market_value}, " \
                              f"委托价格={round(price, 2)}, 持仓可用数量={position.can_use_volume}， 委托数量={size}"
                    log.debug(content)
                    return self.sell(stock_code=stock_code,
                                     volume=size,
                                     price_type=pqconstant.FIX_PRICE,
                                     price=price,
                                     force=force)
                else:
                    content = f"策略名称={self.strategy_name}, 类型=Sell, 股票代码={stock_code}, " \
                              f"目标市值={target}, 持仓市值={market_value}, " \
                              f"委托价格={round(price, 2)},持仓数量={position.volume}, 委托数量={size}, " \
                              f"可用数量（{position.can_use_volume}）不足，放弃交易"
                    log.warning(content)
                    return None  # 未满足交易条件
        else:  # target = market_value
            log.warning(f"调仓目标等于持仓数量，放弃交易")
            return None

    def order_target_percent(self, stock_code: str, target: float = 0.0, force: bool = False):
        total_value = self.portfolio.total_value
        target *= total_value
        return self.order_target_value(stock_code=stock_code, target=target, force=force)

    def order_target_volume(self, stock_code: str, target: float = 0.0, force: bool = False):
        # 获取价格
        # trade_date = self.current_dt.strftime('%Y-%m-%d')
        if self.unit in ['1d']:
            trade_date = self.current_dt.strftime('%Y-%m-%d')
            trade_datetime = None
        else:
            trade_date = None
            trade_datetime = self.current_dt.strftime('%Y-%m-%d %H:%M:%S')

        # price_df = get_history(count=1,
        #                        end_date=trade_date,
        #                        end_datetime=trade_datetime,
        #                        unit=self.unit,
        #                        field='close',
        #                        security_list=[stock_code],
        #                        dividend_type=self.dividend_type,
        #                        expect_df=True)

        price_df = get_factor_v2(limit=1,
                                 end_date=trade_date,
                                 end_datetime=trade_datetime,
                                 timeframe=self.unit,
                                 factor='close',
                                 symbol=[stock_code],
                                 dividend_type=self.dividend_type,
                                 expect_df=True)

        if price_df.empty:
            log.warning(f"股票停牌，取消买入: stock_code={stock_code}, trade_date={trade_date}")
            return
        else:
            price = price_df.iloc[0, 0]

        estimated_target_value = target * price
        return self.order_target_value(stock_code=stock_code, target=estimated_target_value, force=force)

    def close(self, stock_code: str, force: bool = False):
        """
        以最新价平仓股票代码, 默认平仓数量为可用数量

        :param stock_code:
        :param force:
        :return:
        """
        # 持仓查询
        position: Position = self.portfolio.positions.get(stock_code, None)
        volume = position.can_use_volume

        return self.sell(stock_code=stock_code, volume=volume, force=force)

    @property
    def unit(self):
        """
        unit: 单位时间长度，支持1d、1m，默认为1d
        :return:
        """
        return self.__unit

    @property
    def strategy_name(self):
        return self.__strategy_name

    @property
    def index(self):
        return self.__index

    @property
    def stock_pool(self):
        return self.__stock_pool

    @property
    def dividend_type(self):
        """
        dividend_type: 复权选项(对股票/基金的价格字段、成交量字段及factor字段生效)
                'front'
                : 前复权, 默认是前复权
                none
                : 不复权, 返回实际价格
                'back'
                : 后复权
        :return:
        """
        return self.__dividend_type

    @property
    def parameters(self) -> dict:
        """
        执行策略的参数字典
        :return:
        """
        return self.__parameters

    @property
    def current_dt(self) -> datetime.datetime:
        return self.__dt

    def buy(self,
            stock_code: str,
            volume: int,
            price_type: int = pqconstant.FIX_PRICE,
            price: float = None,
            force: bool = False):
        """

        :param stock_code: 股票代码
        :param volume: 数量
        :param price_type: 价格类型
        :param price: 买入价格
        :param force: 如果force == True, 则直接交易，忽略是否进行委托队列。仅在实盘中有效, 在回测中忽略
        :return:
        """
        # 检查股票是否已经停牌
        stock_info = get_stock_info_v2(symbols=[stock_code],
                                       fields=['symbol', 'suspend_status'],
                                       trade_date=self.current_dt.strftime('%Y-%m-%d'))
        if stock_info.empty:
            log.warning(f"无法获取股票信息，取消买入: stock_code={stock_code}, trade_date={self.current_dt.strftime('%Y-%m-%d')}")
            return

        # 停复牌状态：<=0 正常，-1 复牌，>0 停牌, >=1 停牌天数
        suspend_status = stock_info.iloc[0, 1]
        if suspend_status > 0:
            log.warning(
                f"股票停牌，取消买入: stock_code={stock_code}, trade_date={self.current_dt.strftime('%Y-%m-%d')}")
            return

        # 获取价格
        if price_type != pqconstant.FIX_PRICE or price is None:
            # trade_date = self.current_dt.strftime('%Y-%m-%d')
            if self.unit in ['1d']:
                trade_date = self.current_dt.strftime('%Y-%m-%d')
                trade_datetime = None
            else:
                trade_date = None
                trade_datetime = self.current_dt.strftime('%Y-%m-%d %H:%M:%S')
            # price_df = get_history(count=1,
            #                        end_date=trade_date,
            #                        end_datetime=trade_datetime,
            #                        unit=self.unit,
            #                        field='close',
            #                        security_list=[stock_code],
            #                        dividend_type=self.dividend_type,
            #                        expect_df=True)

            # TODO 使用get_factor_v2()获取最新价格
            price_df = get_factor_v2(limit=1,
                                     end_date=trade_date,
                                     end_datetime=trade_datetime,
                                     timeframe=self.unit,
                                     factor='close',
                                     symbol=stock_code,
                                     dividend_type=self.dividend_type,
                                     expect_df=True)

            if price_df.empty:
                log.warning(f"无法获取股票价格，取消买入: stock_code={stock_code}, trade_date={trade_date}")
                return
            else:
                price = price_df.iloc[0, 0]

        # 更新Portfolio cash
        cost = volume * price
        if self.portfolio.available_cash < cost:
            log.warning(f"可用现金不足，放弃买入. cash={self.portfolio.available_cash}, cost={cost}")
            return
        else:
            self.portfolio.add_cash(cost * -1)

        # 更新Portfolio Positions
        position: Position = self.portfolio.positions.get(stock_code, None)
        if position:
            # 已经有头寸
            position.update_position(volume=volume, price=price, transact_time=self.current_dt)
        else:
            # 创建新的头寸
            position = Position(stock_code=stock_code,
                                volume=volume,
                                price=price,
                                init_time=self.current_dt,
                                context=self)
            self.portfolio.positions[stock_code] = position

        # 更新委托Id
        self.order_id += 1
        order = Order(order_id=self.order_id, stock_code=stock_code, direction='buy', order_volume=volume,
                      order_price=price, avg_cost=price, context=self)
        self.notify_order(order)
        return order

    def sell(self,
             stock_code: str,
             volume: int,
             price_type: int = pqconstant.FIX_PRICE,
             price: float = None,
             force: bool = False):
        """
        卖出股票
        :param stock_code:
        :param volume:
        :param price_type:
        :param price:
        :param force:
        :return:
        """
        # 检查股票是否已经停牌
        stock_info = get_stock_info_v2(symbols=[stock_code],
                                       fields=['symbol', 'suspend_status'],
                                       trade_date=self.current_dt.strftime('%Y-%m-%d'))
        if stock_info.empty:
            log.warning(f"无法获取股票信息，取消买入: stock_code={stock_code}, trade_date={self.current_dt.strftime('%Y-%m-%d')}")
            return

        # 停复牌状态：<=0 正常，-1 复牌，>0 停牌, >=1 停牌天数
        suspend_status = stock_info.iloc[0, 1]
        if suspend_status > 0:
            log.warning(
                f"股票停牌，取消买入: stock_code={stock_code}, trade_date={self.current_dt.strftime('%Y-%m-%d')}")
            return

        # 获取价格
        if price_type != pqconstant.FIX_PRICE or price is None:
            # trade_date = self.current_dt.strftime('%Y-%m-%d')
            if self.unit in ['1d']:
                trade_date = self.current_dt.strftime('%Y-%m-%d')
                trade_datetime = None
            else:
                trade_date = None
                trade_datetime = self.current_dt.strftime('%Y-%m-%d %H:%M:%S')
            # price_df = get_history(count=1,
            #                        end_date=trade_date,
            #                        end_datetime=trade_datetime,
            #                        unit=self.unit,
            #                        field='close',
            #                        security_list=[stock_code],
            #                        dividend_type=self.dividend_type,
            #                        expect_df=True)

            # TODO 使用get_factor_v2()获取最新价格
            price_df = get_factor_v2(limit=1,
                                     end_date=trade_date,
                                     end_datetime=trade_datetime,
                                     timeframe=self.unit,
                                     factor='close',
                                     symbol=stock_code,
                                     dividend_type=self.dividend_type,
                                     expect_df=True)

            if price_df.empty:
                log.warning(f"无法获取股票价格，取消卖出: stock_code={stock_code}, trade_date={trade_date}")
                return
            else:
                price = price_df.iloc[0, 0]

        position: Position = self.portfolio.positions.get(stock_code, None)
        if position is None:
            log.error(f"stock_code: {stock_code} 无持仓，取消卖出")
            return
        if position.can_use_volume < volume:
            log.error(f"stock_code: {stock_code} 持仓数量 {position.can_use_volume} 不足，取消卖出")
            return

        # 持仓成本
        avg_price = position.avg_price

        # 更新持仓
        update_status = position.update_position(volume=volume * -1, price=price, transact_time=self.current_dt)

        # 检查是否已经平仓, 如是，则删除头寸记录
        position: Position = self.portfolio.positions.get(stock_code, None)
        if position and position.volume == 0:
            del self.portfolio.positions[stock_code]

        # 更新Portfolio cash
        if update_status:
            self.portfolio.add_cash(volume * price)

        # 更新委托Id
        self.order_id += 1
        order = Order(order_id=self.order_id, stock_code=stock_code, direction='sell', order_volume=volume,
                      order_price=price, avg_cost=avg_price, context=self)
        self.notify_order(order)
        return order

    @property
    def portfolio(self):
        """
        策略投资组合，可通过该对象获取当前策略账户、持仓等信息
        """
        return self._portfolio

    def inout_cash(self, cash: float):
        """
        投资组合转入或转出资金，当日的出入金从当日开始记入成本，用于计算收益，即当日结束计算收益时的本金是包含当日出入金金额的
        :param cash: 可正可负，正为入金，负为出金。
        :return:
        """
        # 记录出入金历史
        self.inout_cash_his.append({"datetime": self.current_dt, "cash": cash})
        # 把出入金更新到投资组合的可以资金
        self.portfolio.add_cash(cash=cash)



import importlib
import importlib.util
import six


def check_module(module_name: str):
    """
    Checks if module can be imported without actually
    importing it
    """

    module_spec = importlib.util.find_spec(module_name)
    if module_spec is None:
        print("Module: {} not found".format(module_name))
        return None
    else:
        return module_spec


def import_module(module_name: str):
    """
    动态导入模块

    :param file_path: 模块文件路径，i.e. strats/strat_demo.py 或者 strats\\strat_demo.py
    :return:
    """
    return importlib.import_module(module_name)


def import_module_from_file(file_path: str):
    """
    动态导入模块

    :param file_path: 模块文件路径，i.e. strats/strat_demo.py 或者 strats\\strat_demo.py
    :return:
    """
    module_name = file_path[:-3]
    module_name = module_name.replace('/', '.').replace('\\', '.')
    return importlib.import_module(module_name)


def check_module_from_file(file_path: str):
    """
    Checks if module can be imported without actually
    importing it
    """
    module_name = file_path[:-3]
    module_name = module_name.replace('/', '.').replace('\\', '.')

    module_spec = importlib.util.find_spec(module_name)
    if module_spec is None:
        print("Module: {} not found".format(module_name))
        return None
    else:
        return module_spec


def import_module_from_spec(module_spec):
    """
    Import the module via the passed in module specification
    Returns the newly imported module
    """
    module = importlib.util.module_from_spec(module_spec)
    module_spec.loader.exec_module(module)
    return module


def import_module_from_code(code: str, kwargs: dict = None):
    """
    从字符串的代码中导入模块
    :param kwargs: 传入策略代码的相关信息
    :param code: 字符串代码
    :return: 代码模块
    """
    if kwargs is not None and 'origin' in kwargs:
        origin = kwargs['origin']
    else:
        origin = 'strategy.py'

    # 从字符串中创建一个ModuleSpec对象
    obj_spec = importlib.util.spec_from_loader('obj_spec',
                                               loader=None,
                                               origin=origin,
                                               is_package=False)
    # 创建一个新模块对象
    obj = importlib.util.module_from_spec(obj_spec)
    # 将Python代码作为字符串传递给exec()函数来执行它
    six.exec_(code, obj.__dict__)
    # 执行新模块中的函数或变量
    return obj


if __name__ == '__main__':
    # case 1: fake module
    file_path = 'strats/hello_world.py'
    module_spec = check_module_from_file(file_path)
    if not module_spec:
        print("fake module: {}".format(file_path))

    # case 2: true module, check before imported
    file_path = 'strats/new_strat_demo.py'
    module_spec = check_module_from_file(file_path=file_path)
    if module_spec:
        strategy = import_module_from_spec(module_spec=module_spec)
        strategy.handle_bar(None)

    # case 3: import module directly
    file_path = 'strats/new_strat_demo.py'
    strategy = import_module_from_file(file_path)
    if hasattr(strategy, 'before_trading_start'):
        print("has before_trading_start")
    strategy.handle_bar(None)

    code = """
import pandas as pd
def initialize():
    print("I'm in Init! imported from code")
"""
    strat = import_module_from_code(code=code)
    strat.initialize()
